@@ -7,21 +7,22 @@ from __future__ import annotations
 import logging
 import os
 import param
-import re
 from enum import Enum, unique
 from param import Parameterized
 from pathlib import Path
-from typing import List, Optional
-from urllib.parse import urlparse
+from typing import Any, Dict, List, Optional
 
 from azureml.train.hyperdrive import HyperDriveConfig
 
 from health_azure import create_crossval_hyperdrive_config
-from health_azure.himl import create_grid_hyperdrive_config
-from health_azure.utils import (RUN_CONTEXT, PathOrString, is_global_rank_zero, is_running_in_azure_ml,
-                                get_amlt_aml_working_dir, is_amulet_job, ENV_AMLT_PROJECT_NAME, ENV_AMLT_INPUT_OUTPUT,
-                                ENV_AMLT_SNAPSHOT_DIR, ENV_AMLT_AZ_BATCHAI_DIR)
+from health_azure.himl import (create_grid_hyperdrive_config, create_crossval_hyperparam_args_v2,
+                               create_grid_hyperparam_args_v2)
+from health_azure.amulet import (ENV_AMLT_PROJECT_NAME, ENV_AMLT_INPUT_OUTPUT,
+                                 ENV_AMLT_SNAPSHOT_DIR, ENV_AMLT_AZ_BATCHAI_DIR,
+                                 is_amulet_job, get_amulet_aml_working_dir)
+from health_azure.utils import (RUN_CONTEXT, PathOrString, is_global_rank_zero, is_running_in_azure_ml)
 from health_ml.utils import fixed_paths
+from health_ml.utils.checkpoint_utils import CheckpointParser
 from health_ml.utils.common_utils import (CHECKPOINT_FOLDER,
                                           create_unique_timestamp_id,
                                           DEFAULT_AML_UPLOAD_DIR,
@@ -121,7 +122,7 @@ class ExperimentFolderHandler(Parameterized):
                 # Job submitted via Amulet
                 amlt_root_folder = Path(os.environ[ENV_AMLT_INPUT_OUTPUT])
                 project_name = os.environ[ENV_AMLT_PROJECT_NAME]
-                snapshot_dir = get_amlt_aml_working_dir()
+                snapshot_dir = get_amulet_aml_working_dir()
                 assert snapshot_dir, \
                     f"Either {ENV_AMLT_SNAPSHOT_DIR} or {ENV_AMLT_AZ_BATCHAI_DIR} must exist in env vars"
                 print(f"Found the following environment variables set by Amulet: "
@@ -131,9 +132,9 @@ class ExperimentFolderHandler(Parameterized):
                 outputs_folder = snapshot_dir / DEFAULT_AML_UPLOAD_DIR
                 logs_folder = snapshot_dir / DEFAULT_LOGS_DIR_NAME
 
-        print(f"Run outputs folder: {outputs_folder}")
-        print(f"Logs folder: {logs_folder}")
-        print(f"Run root directory: {run_folder}")
+        logging.info(f"Run outputs folder: {outputs_folder}")
+        logging.info(f"Logs folder: {logs_folder}")
+        logging.info(f"Run root directory: {run_folder}")
         return ExperimentFolderHandler(
             outputs_folder=outputs_folder,
             logs_folder=logs_folder,
@@ -142,32 +143,13 @@ class ExperimentFolderHandler(Parameterized):
         )
 
 
-SRC_CHECKPOINT_FORMAT_DOC = ("<AzureML_run_id>:<optional/custom/path/to/checkpoints/><filename.ckpt>"
-                             "If no custom path is provided (e.g., <AzureML_run_id>:<filename.ckpt>)"
-                             "the checkpoint will be downloaded from the default checkpoint folder "
-                             "(e.g., 'outputs/checkpoints/'). If no filename is provided, (e.g., "
-                             "`src_checkpoint=<AzureML_run_id>`) the latest checkpoint (last.ckpt) "
-                             "will be used to initialize the model."
-                             )
-
-
 class WorkflowParams(param.Parameterized):
     """
     This class contains all parameters that affect how the whole training and testing workflow is executed.
     """
     random_seed: int = param.Integer(42, doc="The seed to use for all random number generators.")
-    src_checkpoint: str = param.String(default="",
-                                       doc="This flag can be used in 3 different scenarios:"
-                                           "1- Resume training from a checkpoint to train longer."
-                                           "2- Run inference-only using `run_inference_only` flag jointly."
-                                           "3- Transfer learning from a pretrained model checkpoint."
-                                           "We currently support three types of checkpoints: "
-                                           "    a. A local checkpoint folder that contains a checkpoint file."
-                                           "    b. A URL to a remote checkpoint to be downloaded."
-                                           "    c. A previous azureml run id where the checkpoint is supposed to be "
-                                           "       saved ('outputs/checkpoints/' folder by default.)"
-                                           "For the latter case 'c' : src_checkpoint should be in the format of "
-                                           f"{SRC_CHECKPOINT_FORMAT_DOC}")
+    src_checkpoint: CheckpointParser = param.ClassSelector(class_=CheckpointParser, default=None,
+                                                           instantiate=False, doc=CheckpointParser.DOC)
     crossval_count: int = param.Integer(default=1, bounds=(0, None),
                                         doc="The number of splits to use when doing cross-validation. "
                                             "Use 1 to disable cross-validation")
@@ -193,7 +175,9 @@ class WorkflowParams(param.Parameterized):
     regression_metrics: str = param.String(default=None, doc="A list of names of metrics to compare")
     run_inference_only: bool = param.Boolean(False, doc="If True, run only inference and skip training after loading"
                                                         "model weights from the specified checkpoint in "
-                                                        "`src_checkpoint` flag. If False, run training and inference.")
+                                                        "`src_checkpoint` flag. Inference is run on both validation "
+                                                        "and test sets. If False, run training and inference.")
+    resume_training: bool = param.Boolean(False, doc="If True, resume training from the src_checkpoint.")
     tag: str = param.String(doc="A string that will be used as the display name of the run in AzureML.")
     experiment: str = param.String(default="", doc="The name of the AzureML experiment to use for this run. If not "
                                    "provided, the name of the model class will be used.")
@@ -211,42 +195,15 @@ class WorkflowParams(param.Parameterized):
     CROSSVAL_COUNT_ARG_NAME = "crossval_count"
     RANDOM_SEED_ARG_NAME = "random_seed"
 
-    @property
-    def src_checkpoint_is_url(self) -> bool:
-        try:
-            result = urlparse(self.src_checkpoint)
-            return all([result.scheme, result.netloc])
-        except ValueError:
-            return False
-
-    @property
-    def src_checkpoint_is_local_file(self) -> bool:
-        return Path(self.src_checkpoint).is_file()
-
-    @property
-    def src_checkpoint_is_aml_run_id(self) -> bool:
-        match = re.match(r"[_\w-]*$", self.src_checkpoint.split(":")[0])
-        return match is not None and not self.src_checkpoint_is_url and not self.src_checkpoint_is_local_file
-
-    @property
-    def is_valid_src_checkpoint(self) -> bool:
-        if self.src_checkpoint:
-            return self.src_checkpoint_is_local_file or self.src_checkpoint_is_url or self.src_checkpoint_is_aml_run_id
-        return True
-
     def validate(self) -> None:
-        if not self.is_valid_src_checkpoint:
-            raise ValueError(f"Invalid src_checkpoint: {self.src_checkpoint}. Please provide a valid URL, local file "
-                             "or azureml run id.")
         if self.crossval_count > 1:
             if not (0 <= self.crossval_index < self.crossval_count):
                 raise ValueError(f"Attribute crossval_index out of bounds (crossval_count = {self.crossval_count})")
 
         if self.run_inference_only and not self.src_checkpoint:
-            raise ValueError("Cannot run inference without a src_checkpoint. Please specify a valid src_checkpoint."
-                             "You can either use a URL, a local file or an azureml run id. For custom checkpoint paths "
-                             "within an azureml run, (other than last.ckpt), provide a src_checkpoint in the format."
-                             f"{SRC_CHECKPOINT_FORMAT_DOC}")
+            raise ValueError(f"Cannot run inference without a src_checkpoint. {CheckpointParser.INFO_MESSAGE}")
+        if self.resume_training and not self.src_checkpoint:
+            raise ValueError(f"Cannot resume training without a src_checkpoint. {CheckpointParser.INFO_MESSAGE}")
 
     @property
     def is_running_in_aml(self) -> bool:
@@ -294,8 +251,31 @@ class WorkflowParams(param.Parameterized):
                                              metric_name="val/loss"
                                              )
 
+    def get_crossval_hyperparam_args_v2(self) -> Dict[str, Any]:
+        """
+        Wrapper function to create hyperparameter search arguments specifically for running cross validation
+        with AML SDK v2
+
+        :return: A dictionary of hyperparameter search arguments and values.
+        """
+        return create_crossval_hyperparam_args_v2(num_splits=self.crossval_count,
+                                                  cross_val_index_arg_name=self.CROSSVAL_INDEX_ARG_NAME,
+                                                  metric_name="val/loss")
+
+    def get_grid_hyperparam_args_v2(self) -> Dict[str, Any]:
+        """
+        Wrapper function to create hyperparameter search arguments specifically for running grid search
+        with AML SDK v2
+
+        :return: A dictionary of hyperparameter search arguments and values.
+        """
+        return create_grid_hyperparam_args_v2(values=list(map(str, range(self.different_seeds))),
+                                              argument_name=self.RANDOM_SEED_ARG_NAME,
+                                              metric_name="val/loss")
+
 
 class DatasetParams(param.Parameterized):
+    datastore: str = param.String(default="", doc="Datastore to look for data in")
     azure_datasets: List[str] = param.List(default=[], class_=str,
                                            doc="If provided, the ID of one or more datasets to use when running in"
                                                " AzureML. This dataset must exist as a folder of the same name "
@@ -508,18 +488,37 @@ class TrainerParams(param.Parameterized):
         param.String(default=None,
                      doc="The value to use for the 'profiler' argument for the Lightning trainer. "
                          "Set to either 'simple', 'advanced', or 'pytorch'")
+    pl_sync_batchnorm: bool = param.Boolean(default=True,
+                                            doc="PyTorch Lightning trainer flag 'sync_batchnorm': If True, "
+                                            "synchronize batchnorm across all GPUs when running in ddp mode."
+                                            "If False, batchnorm is not synchronized.")
     monitor_gpu: bool = param.Boolean(default=False,
                                       doc="If True, add the GPUStatsMonitor callback to the Lightning trainer object. "
                                           "This will write GPU utilization metrics every 50 batches by default.")
     monitor_loading: bool = param.Boolean(default=False,
                                           doc="If True, add the BatchTimeCallback callback to the Lightning trainer "
                                               "object. This will monitor how long individual batches take to load.")
+    monitor_training: bool = param.Boolean(default=False,
+                                           doc="If True, add the TrainingDiagnosisCallback to the Lightning trainer "
+                                               "object. This will monitor when training, validation and test starts "
+                                               "and ends and also intermediate steps.")
     run_extra_val_epoch: bool = param.Boolean(default=False,
                                               doc="If True, run an additional validation epoch at the end of training "
                                               "to produce plots outputs on the validation set. This is to reduce "
                                               "any validation overheads during training time and produce "
                                               "additional time or memory consuming outputs only once after "
                                               "training is finished on the validation set.")
+    pl_accumulate_grad_batches: int = param.Integer(default=1,
+                                                    doc="The number of batches over which gradients are accumulated, "
+                                                    "before a parameter update is done.")
+    pl_log_every_n_steps: int = param.Integer(default=50,
+                                              doc="PyTorch Lightning trainer flag 'log_every_n_steps': How often to"
+                                              "log within steps. Default to 50.")
+    pl_replace_sampler_ddp: bool = param.Boolean(default=True,
+                                                 doc="PyTorch Lightning trainer flag 'replace_sampler_ddp' that "
+                                                 "sets the sampler for distributed training with shuffle=True during "
+                                                 "training and shuffle=False during validation. Default to True. Set to"
+                                                 "False to set your own sampler.")
 
     @property
     def use_gpu(self) -> bool:
